package org.icmvc.nvr.dispatcher.wfm.camel;

/*
 * ###license-information-start###
 * Toolbox for Research - TTP Dispatcher
 * __
 * Copyright (C) 2017 The MOSAIC Project - Institut fuer Community Medicine der
 * 							Universitaetsmedizin Greifswald - mosaic@uni-greifswald.de
 * __
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * ###license-information-end###
 */

import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import javax.ejb.EJB;
import javax.inject.Named;

import org.apache.camel.Exchange;
import org.apache.camel.LoggingLevel;
import org.apache.camel.Message;
import org.apache.camel.Processor;
import org.apache.camel.builder.RouteBuilder;
import org.apache.camel.component.properties.PropertiesComponent;
import org.apache.camel.processor.aggregate.AggregationStrategy;
import org.apache.http.NameValuePair;
import org.apache.http.client.entity.UrlEncodedFormEntity;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.cookie.Cookie;
import org.apache.http.impl.client.BasicCookieStore;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClients;
import org.apache.http.message.BasicNameValuePair;
import org.apache.http.util.EntityUtils;
import org.icmvc.commons.audit.interfaces.IAudit;
import org.icmvc.ttp.dispatcher.common.config.DispatcherConfig;
import org.icmvc.ttp.dispatcher.common.exception.ValidationException;
import org.icmvc.ttp.dispatcher.common.model.dto.DispatcherDto.DtoType;
import org.icmvc.ttp.dispatcher.common.services.IConfigManager;
import org.icmvc.ttp.dispatcher.common.workflow.ServiceWorkflows.PsnWorkflows;
import org.icmvc.ttp.nvr.dispatcher.common.Constants.DispatcherExtData;
import org.icmvc.ttp.nvr.dispatcher.common.Workflows;
import org.icmvc.ttp.nvr.dispatcher.common.interfaces.IPropertiesDbResolveService;
import org.icmvc.ttp.nvr.dispatcher.common.model.OpenClinicaLoginModel;
import org.icmvc.ttp.nvr.dispatcher.common.model.SiteModel;
import org.jsoup.Jsoup;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 * @author Peter Penndorf
 *
 */
@Named
public class Workflow extends RouteBuilder
{
	private final Logger logger = LoggerFactory.getLogger(this.getClass());

	@EJB
	private volatile IConfigManager config;
	@EJB
	private volatile IAudit audit;
	@EJB
	private IPropertiesDbResolveService propResolver;
	private volatile DispatcherConfig dConf;
	private volatile String studienId;

	private void init() throws ValidationException
	{
		dConf = config.getDispatcherConfig();
		PropertiesComponent c = (PropertiesComponent) getContext().getComponent("properties");
		PropertiesDbResolveFunction propResolverFuntion = new PropertiesDbResolveFunction(propResolver);
		c.addFunction(propResolverFuntion);
		studienId = propResolverFuntion.apply("oc.study.default.id");
	}

	@Override
	public void configure() throws Exception
	{
		init();
		logger.debug("StudienID: {}", studienId);
		// @formatter:off

		// final DefaultErrorHandlerBuilder dlqBuilder =
		// deadLetterChannel("direct:limsPsnDLQ")
		// .delayPattern("1:5000;2:30000;5:60000")
		// .maximumRedeliveries(0);

		// ### Camel Interception
		// intercept().to("log:interception");
		// ### ErroHandling
		errorHandler(noErrorHandler());

		// ### Helper-Routes
		from("direct:setCommonHeader")
			.setHeader("usernameToken", simple("{{dbprop:oc.usernameToken}}"))
			.setHeader("username", simple("{{dbprop:oc.username}}"))
			.setHeader("password", simple("{{dbprop:oc.password}}"))
			.setHeader("eventId", simple("{{dbprop:oc.event.default.id}}"))
			.setHeader("enrDate", simple("date:now:yyyy-MM-dd"))
			.setHeader("startDate", simple("date:now:yyyy-MM-dd"))
			.setHeader("startTime", simple("date:now:hh:MM"))
			.setHeader("studyId", simple("{{dbprop:oc.study.default.id}}"));

		from("direct:checkServiceResponse").convertBodyTo(String.class, "UTF-8")
//			.log(LoggingLevel.DEBUG, logger, "checking response '${body}'")
			.setHeader("responseFaultString", xpath("//*[local-name()='faultstring']/text()"))
			.setHeader("responseResult", xpath("//*[local-name()='result']/text()"))
			.choice().when(header("responseFaultString").isEqualTo(null))
				.log(LoggingLevel.ERROR, logger, "Service returned error meassage '${header.responseFaultString}'")
				.throwException(new Exception())
			.otherwise().when(header("responseResult").isNotEqualTo("Success"))
				.setHeader("responseMessage", xpath("//*[local-name()='error']/text()"))
				.log(LoggingLevel.ERROR, logger, "Call of Service was not successfully. Message: ${header.responseMessage}")
				.throwException(new Exception());

		// ### Main-Routes
		from(Workflows.CHECK_LOGGED_IN)
			.setHeader(Exchange.HTTP_METHOD, constant("GET"))
			.setHeader("COOKIE", simple("JSESSIONID=${body}; path=/OpenClinica"))
			.to("http4:{{dbprop:oc.url.checkLoggedIn}}")
			.convertBodyTo(String.class, "UTF-8")
			.choice().when(body().contains("loginBox"))
				.log(LoggingLevel.DEBUG, logger, constant("### not logged in").toString())
				.setBody(constant(false))
			.otherwise()
				.log(LoggingLevel.DEBUG, logger, constant("### logged in").toString())
			.setBody(constant(true)).end();

		from(Workflows.CREATE_PATIENT)
		// Generate PSNs
		.to(PsnWorkflows.GET_ID_FROM_ID)
		// Set DTO in Header to preserve it during the next steps
		.setHeader("dispatcherDTO", body())
		.setHeader("psn", simple("body.getDTOByString('" + DtoType.SEQUENCE_PSN + "').getReturnValue()", String.class))
		// When no PSN could be generated
		.choice().when(header("psn").isNull())
			.throwException(new Exception("Could not generate PSN"))
		.otherwise()
			.setHeader("birthdate", simple("body.extData.get('"+DispatcherExtData.BIRTHDATE+"')", String.class))
			.setHeader("siteId", simple("body.extData.get('"+DispatcherExtData.SITE+"')", String.class))
			.setHeader("gender", simple("body.extData.get('"+DispatcherExtData.GENDER+"')", String.class))
			.to("direct:setCommonHeader")
			// Check if subject is already in Study
			.to("freemarker:soap_templates/isStudySubject.xml").convertBodyTo(String.class, "UTF-8")
			.setHeader(Exchange.HTTP_METHOD, simple("POST"))
			.setHeader(Exchange.CONTENT_TYPE, constant("text/xml"))
			.to("http4:{{dbprop:oc.url.studySubjectService}}")
			.setHeader("responseResult", xpath("//*[local-name()='result']/text()"))
			.log(LoggingLevel.TRACE, logger, simple("isStudySubject ? ${headers.responseResult}").toString())
			// If subject does not exists, create and register
			.choice().when(header("responseResult").isEqualTo("Fail"))
				.log(LoggingLevel.DEBUG, logger, constant("### could not find subject with id '${header.psn}' in site '${header.siteId}', creating one").toString())
				.to("freemarker:soap_templates/createPatient.xml").convertBodyTo(String.class, "UTF-8")
				.setHeader(Exchange.HTTP_METHOD, simple("POST"))
				.setHeader(Exchange.CONTENT_TYPE, constant("text/xml"))
				.to("http4:{{dbprop:oc.url.studySubjectService}}")
				.to("direct:checkServiceResponse")
				.to("freemarker:soap_templates/registerEvent.xml").convertBodyTo(String.class, "UTF-8")
				.setHeader(Exchange.HTTP_METHOD, simple("POST"))
				.setHeader(Exchange.CONTENT_TYPE, constant("text/xml"))
				.to("http4:{{dbprop:oc.url.registerEvent}}")
				.to("direct:checkServiceResponse")
			.end()
		// Restore DispatcherDTO to body
		.setBody(header("dispatcherDTO"));

		from(Workflows.GET_SITES)
			.to("direct:setCommonHeader")
			.to("freemarker:soap_templates/getSitesList.xml").convertBodyTo(String.class, "UTF-8")
			.setHeader(Exchange.HTTP_METHOD, simple("POST"))
			.setHeader(Exchange.CONTENT_TYPE, constant("text/xml"))
			.to("http4:{{dbprop:oc.url.getSites}}")
			.to("direct:checkServiceResponse")
			.to("direct:parseSites");

		from("direct:parseSites")
			.setBody(xpath("//*[local-name()='study'][*[local-name()='identifier']='"+studienId+"']"))
			.split(body().tokenizeXML("site", ""), new AggregationStrategy()
			{
				@Override
				public Exchange aggregate(Exchange oldExchange, Exchange newExchange)
				{
					if (oldExchange == null)
					{
						Message m = newExchange.getIn();
						SiteModel s = new SiteModel(m.getHeader("id").toString(), m.getHeader("name").toString(), m.getHeader("oid").toString());
						logger.debug(s.toString());
						newExchange.getIn().setBody(new ArrayList<SiteModel>());
						((ArrayList<SiteModel>) newExchange.getIn().getBody()).add(s);
						return newExchange;
					}

					Message m = newExchange.getIn();
					SiteModel s = new SiteModel(m.getHeader("id").toString(), m.getHeader("name").toString(), m.getHeader("oid").toString());
					((ArrayList<SiteModel>) oldExchange.getIn().getBody()).add(s);

					logger.debug(s.toString());

					return oldExchange;
				}
			})
//			.log(LoggingLevel.DEBUG, logger, simple("${body}").toString())
			.setHeader("id", xpath("//*[local-name()='identifier']/text()").stringResult())
			.setHeader("oid", xpath("//*[local-name()='oid']/text()").stringResult())
			.setHeader("name", xpath("//*[local-name()='name']/text()").stringResult());

		from(Workflows.LOG_IN)
			.setHeader("loginUrl", simple("http:{{dbprop:oc.url.logIn}}"))
			.setHeader("logoutUrl", simple("http:{{dbprop:oc.url.logOut}}"))
//			.setBody(constant(new OpenClinicaLoginModel("root", "openclinica")))
			.process(new Processor()
			{
				@Override
				public void process(Exchange exchange) throws Exception
				{
					Message in = exchange.getIn();
					OpenClinicaLoginModel m = (OpenClinicaLoginModel) in.getBody();
					BasicCookieStore store = new BasicCookieStore();
					CloseableHttpClient httpclient = HttpClients.custom().setDefaultCookieStore(store).build();
					HttpPost httpPost = new HttpPost(in.getHeader("loginUrl").toString());
					List<NameValuePair> nvps = new ArrayList<NameValuePair>();
					nvps.add(new BasicNameValuePair("j_username", m.getUsername()));
					nvps.add(new BasicNameValuePair("j_password", m.getPassword()));
					nvps.add(new BasicNameValuePair("submit", "Login"));
					httpPost.setEntity(new UrlEncodedFormEntity(nvps));
					CloseableHttpResponse response = httpclient.execute(httpPost);
					if(response.getStatusLine().getStatusCode()==302 && !response.getHeaders("Location")[0].getValue().contains("error"))
					{
						for(Cookie c: store.getCookies())
						{
							javax.servlet.http.Cookie sc = new javax.servlet.http.Cookie(c.getName(), c.getValue());
//							sc.setDomain(c.getDomain());
//							sc.setPath(c.getPath().substring(0, c.getPath().lastIndexOf("/")));
							sc.setPath(c.getPath());
							if(c.getExpiryDate()!=null)
							{
								sc.setMaxAge((int) ((c.getExpiryDate().getTime() - (new Date().getTime()))/1000));
							}
							m.getCookies().add(sc);
							logger.debug("stored cookie '{}' with path '{}' and value '{}'", sc.getName(), sc.getPath(), sc.getValue());
						}
						m.setLoggedIn(true);

						HttpGet httpGet = new HttpGet(response.getHeaders("Location")[0].getValue().split(";")[0]);
						logger.debug("Location of response: {}",response.getHeaders("Location")[0].getValue().split(";")[0]);
						CloseableHttpResponse responseGet = httpclient.execute(httpGet);
						if(responseGet.getStatusLine().getStatusCode()==200)
						{
							m.setSite(parseLoginResponseForSite(EntityUtils.toString(responseGet.getEntity())));
						}
						else
						{
							logger.warn("response after login was {}",responseGet.getStatusLine().getStatusCode());
						}
						responseGet.close();
					}
					else
					{
						m.setLoggedIn(false);
					}
					response.close();
					in.setBody(m);

					//Logout
//					HttpGet httpGet = new HttpGet(in.getHeader("logoutUrl").toString());
//					response = httpclient.execute(httpGet);
//					logger.debug("logout-StatusCode = {}", response.getStatusLine().getStatusCode());
//					response.close();

					httpclient.close();
				}

				private String parseLoginResponseForSite(String input)
				{
					try
					{
//						logger.debug("Html-Response:\n{}",input);
//						DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
						org.jsoup.nodes.Document d = Jsoup.parse(input);
//						factory.setNamespaceAware(false);
//						factory.setValidating(false);
//						factory.setExpandEntityReferences(false);
//						Document d = factory.newDocumentBuilder().parse(input);
						org.jsoup.nodes.Element studyElement = d.getElementById("StudyInfo");
						String content = studyElement.text();
						logger.debug("content of studyInfo\n {}", content);
						String site = content.substring(content.indexOf("(")+1, content.indexOf(")"));
						logger.debug("parsed site: {}", site);
						if(site==null || propResolver.getProperty(IPropertiesDbResolveService.COMMON_NVR,"oc.study.default.id").toString().equalsIgnoreCase(site.trim()))
						{
							return null;
						} else
						{
							return site.trim();
						}
					} catch (Exception e)
					{
						logger.error(e.getMessage(), e);
					}
					return null;
				}
			})
			.log(simple("LoggedIn ? ${in.body.isLoggedIn}").toString());

		//@formatter:on
	}

}
